/**************************************************************************
**   yapoo 2010
**************************************************************************/
#include "TsPacketReader.h"
#include <cassert>
#include <stdint.h>
//#include "table.hpp"
//#include "psisection.h"

#include <fstream>

#include <boost/crc.hpp>
//#include <boost/integer.hpp>
//#include <cstddef>
boost::crc_optimal<32, 0x04C11DB7, 0xFFFFFFFF, 0, false, false > crc32;

#include <iostream>
using namespace std;

#define MIN(x,y) ((y<x?y:x))
#define SECT_TID(p) (*p)

//#define SECT_LEN(p) ( ( ((((p)[1])&0x0f) << 8) + (p)[2] + 3 ) )
inline int SECT_LEN( const uint8_t *p ) { return ((p[1]&0x0f) << 8) + p[2] + 3; }


int TsPacketReader::checkCrc( const uint8_t* sectHead, int len )
{
	const uint8_t* crcPos = &sectHead[len - 4];
	uint32_t expected = ( crcPos[0] << 24 ) | ( crcPos[1] << 16 ) | ( crcPos[2] << 8 ) |  crcPos[3];

	crc32.reset();
	crc32.process_block( sectHead, crcPos );
	uint32_t sum = crc32.checksum();

	if ( sum != expected )
	{
#if 0
ofstream ofs( "dump" );
ofs.write( (char*)sectHead, len );
#endif
		return -1;
	}
	return 0;
}


int TsPacketReader::push( const uint8_t *packet )
{
	int ret;

	if ( ! (status == no_data) && ! (status == partial) )
	{
		return ERR_BUFFER;
	}

	tsh.setPos( packet );

#if 1
	if ( !tsh.syncOK() )
	{
assert( tsh.syncOK() );
		return IGN_SYNC;
	}
#endif
	if ( tsh.isError() )
	{
		return IGN_ERRORBIT;
	}
	if ( !tsh.hasData() )
	{
		return IGN_NODATA;
	}

	this->packet = packet;

	if ( status == no_data )
	{
		assert( writer.getPos() == 0 );
		if ( !tsh.isStart() )
		{
			return IGN_START;
		}
		counter = tsh.counter();
		const uint8_t *p = tsh.siHead();
		nextPosInPacket = p - packet;
		return doCopy( SECT_LEN(p) );
	}

	assert( status == partial );
	assert( writer.getPos() > 3 );

	ret = checkCounter();
	if ( ret )
	{
		if ( ret == -1 )
		{
			return IGN_COUNTER;
		}
		else
		{
			return ERR_COUNTER;
		}
	}

	const uint8_t *p = tsh.siNext();
	nextPosInPacket = p - packet;
	return doCopy( SECT_LEN( writer.getBuff() ) );
}

int TsPacketReader::checkCounter()
{
	int next = (counter + 1) & 0x0f;
	if ( next == tsh.counter() )
	{
		counter = next;
		return 0;
	}
	if ( counter == tsh.counter() )
	{
		return -1;
	}
	return -2;
}

int TsPacketReader::next() {
	assert( writer.getPos() == 0 );

	if ( status != no_data )
	{
		return ERR_BUFFER;
	}
	return doCopy( SECT_LEN( &packet[nextPosInPacket] ) );
}


int TsPacketReader::doCopy( int sectionLength )
{
	assert( nextPosInPacket >= 4 ); // TSヘッダ分はスキップしてる
	assert( nextPosInPacket <= 188 ); // 前回最終まで読んだ時は188 それ以上なら内部のバグ

	if ( nextPosInPacket == 188 )
	{
		return BUFFERING;
		// 続けて呼ばれた場合に備え nextPosInPacket は変えない
	}

	const uint8_t *p = &packet[nextPosInPacket];
//	assert( validTable( *p ) || *p == 0xff );

	// バッファ中で無く、テーブルがスタッフィングor不明な場合
	if ( status == no_data &&
		( *p == 0xFF || !validTable( *p ) )
		 ) //validTableはとりあえず実装しないかも
	{
		return BUFFERING;
		// 続けて呼ばれた場合に備え nextPosInPacket は変えない
	}

	// 開始については最低でもセクション長まで(3bytes)まとめて取れない時は拒否
	if ( nextPosInPacket > 185  )
	{
	// status = no_data;
//		return BUFFERING;
		resetStatus();
		return ERR_ERROR;
	}


	int copyLength;
	int sectionLeft = sectionLength - writer.getPos();
	int packetLeft;
	int ret;

	packetLeft = 188 - nextPosInPacket;
	copyLength = MIN( packetLeft, sectionLeft );

	ret = writer.append( p, copyLength );
	if ( ret < 0 )
	{
		resetStatus();
		return ERR_BUFFER;
	}

	nextPosInPacket += copyLength; // ここでposinpacketの範囲チェックはしない

	if ( writer.getPos() == sectionLength )
	{
#if 0
		ret = checkCrc( writer.getBuff(), sectionLength );
		if ( ret )
		{
			resetStatus();
			return ERR_CRC;
		}
#endif
		status = complete;
		return COMPLETE;
	}
	status = partial;
	return BUFFERING;
}
