/**************************************************************************
**   yapoo 2010
**************************************************************************/
#ifndef TSPACKETREADER_H
#define TSPACKETREADER_H

#include <cassert>
#include <stdint.h>
#include <string.h>


class TsHeadParser {
	const uint8_t *packet;

  public:
//	enum { PSI, SI, PES };

	void setPos( const uint8_t* packet ) { this->packet = packet; }
	void setPos( char* packet ) { this->packet = (const uint8_t*)packet; }

	const uint8_t *getPos() { return packet; }

	uint8_t sync() { return packet[0]; }
	bool syncOK() { return (packet[0] == 0x47); }

	uint8_t rawError() { return (packet[1] & 0x80); }
	bool isError() { return rawError(); }
	uint8_t error() { return ( rawError() >> 7 ); }

	uint8_t rawStart() { return (packet[1] & 0x40); }
	bool isStart() { return rawStart(); }

	uint8_t rawPrio() { return (packet[2] & 0x20); }
	bool isPrio() { return rawPrio(); }

	uint8_t rawPidH() { return (packet[1] & 0x1f); }
	uint8_t rawPidL() { return packet[2]; }
	uint16_t pid() { return ( rawPidH() << 8 ) | rawPidL(); }

	uint8_t rawTsc() { return (packet[3] & 0xc0); }
	uint8_t tsc() { return (rawTsc() >> 6); }

	uint8_t rawAfc() { return (packet[3] & 0x30); }
	uint8_t afc() { return (rawAfc() >> 4); }
	bool hasAf() { return (packet[3] & 0x20); }
	bool hasData() { return (packet[3] & 0x10); }

	uint8_t counter() { return (packet[3] & 0x0f); } // memo アダプテーションだけの時は増えない

	uint8_t pointerField() // isStart()なPSI/SIの場合のみ有効
	{
		if ( hasAf() ) {
			// assert()
			return packet[5 + afLen()];
		} else {
			return packet[4];
		}
	}

	unsigned int afLen() { return packet[4]; } //hasAf()な場合


	const uint8_t *siNext() // 前パケットの続き分
	{
		const uint8_t *ret;

		if ( hasAf() ){
			ret = &packet[5 + afLen()];
		} else {
			ret = &packet[4];
		}

		if (isStart() ) {
			ret ++;
		}

		return ret;
	}

	const uint8_t *siHead() // セクション開始位置　（isStart() な時のみ有効）
	{
		const uint8_t *ret;

		if ( hasAf() ){
			ret = &packet[5 + afLen()];
		} else {
			ret = &packet[5];
		}

		ret += pointerField();

		return ret;
	}

	const uint8_t *pesHead() // PES開始位置　（isStart() な時のみ有効）
	{
		const uint8_t *ret;

		if ( hasAf() ){
			ret = &packet[4 + afLen()];
		} else {
			ret = &packet[4];
		}

		return ret;
	}

	const uint8_t *pesNext() // 前パケットの続き分
	{
		const uint8_t *ret;

		if ( hasAf() ){
			ret = &packet[5 + afLen()];
		} else {
			ret = &packet[4];
		}

		return ret;
	}

};


struct SiBuffer
{
public:
	uint8_t *pBuff;
	int currentLength;
	const int bufferSize;
	SiBuffer(uint8_t* p, int n):pBuff(p), bufferSize(n) {}
//	virtual bool validTid( uint8_t tid ) = 0;
};

struct SiBuffer1024 : SiBuffer
{
	SiBuffer1024() : SiBuffer( buff, 1024 ) {};
	uint8_t buff[1024];
};

struct SiBuffer4096 : SiBuffer
{
	SiBuffer4096() : SiBuffer( buff, 4096 ) {};
	uint8_t buff[4096];
};



class TsPacketReader {
public:

	enum ReturnStatus { COMPLETE = 0x0, BUFFERING = 1,
						IGN_ANY = 0x100, IGN_SYNC = 0x0101, IGN_START, IGN_COUNTER, IGN_NODATA, IGN_ERRORBIT,
	//					RST_ANY = 0x1000, RST_COUNTER = 0x1001,
						ERR_ANY = 0x10000, /*ERR_CRC = 0x10001,*/ ERR_COUNTER = 0x10002, ERR_BUFFER = 0x10003, ERR_ERROR };

protected:
	TsHeadParser tsh;
	int counter;
	int nextPosInPacket;

	enum { no_data = -1 /*, partial_no_header */, partial = 1, complete = 0, buffer_error = 0x10000 } readingStatus;
	int status;

	class {
		SiBuffer* pBuffer;

		public:
		int getPos() { return pBuffer->currentLength; }
		uint8_t* getBuff() { return pBuffer->pBuff; }
		SiBuffer* getContainer() { return pBuffer; }

		void setBuffer( SiBuffer *p )
		{
			pBuffer = p;
			pBuffer->currentLength = 0;
		}
		void clearBuffer() { pBuffer->currentLength = 0; }

		int append( const uint8_t* p, int size )
		{
			int n;
			n = pBuffer->currentLength + size;
			if ( n > pBuffer->bufferSize )
			{
				return -1;
			}
			::memcpy( &pBuffer->pBuff[pBuffer->currentLength], p, size );
			pBuffer->currentLength += size;
			assert( pBuffer->currentLength >= 3 ); // ARIB仕様では無いはず（最低でセクションヘッダー分）
			return 0;
		}

	} writer;

	const uint8_t* packet;

	void resetStatus() // 次のpush()以降正常に処理出来る状態に
	{
//		nextPosInPacket = 0;
		writer.clearBuffer(); // pBufferのチェックはしてない
		status = no_data;
	}

	int checkCounter();
	bool validTable( uint8_t ) { return true; }
	int doCopy( int sectionLength );

public:
	TsPacketReader()
	{
		nextPosInPacket = 200; // 不正値にしとく
		status = buffer_error;
	};

	TsPacketReader( SiBuffer *pBuffer )
	{
		setBuffer( pBuffer );
	};

	void clearBuffer()
	{
		writer.clearBuffer();
		resetStatus();
	}

	void setBuffer( SiBuffer *pBuffer )
	{
		writer.setBuffer( pBuffer );
		resetStatus();
	}

	SiBuffer *remindBuffer() { return writer.getContainer(); }

	int push( const uint8_t *packet );
	int push( const char *packet ) { return push( (const uint8_t*)packet ); }

//	int next( SiBuffer *pBuf );
	int next(); // { next( pBuffer ); }

	int checkCrc( const uint8_t* sectHead, int len );
	int checkCrc() { return checkCrc( writer.getBuff(), writer.getPos() ); }
};

#endif // TSPACKETREADER_H
