/**************************************************************************
**   yapoo 2010
**************************************************************************/
#include <stdint.h>
#include <math.h>
#include "aribstring.h"

#define ESC_CHAR 0x1b
#define LS0	0x0f
#define LS1	0x0e
#define APR 0x0d
#define SS2 0x19
#define SS3 0x1d
#define SP  0x20

#define DEL 0x7f
#define MSZ 0x89
#define NSZ 0x8a
#define CSI 0x9b

static const CharType SPACE[] = { 0x3000, ' '};

#define B0(x) (x >> 24)
#define B1(x) ((x >> 16)&0xff)
#define B2(x) ((x >> 8)&0xff)
#define B3(x) (x&0xff)

string Conv::utf8() {
	string ret;

	ret = "";
	StrType::iterator it;
	uint32_t b1,b2,b3,ch;

	it = pStrOut->begin();
	while ( it < pStrOut->end() ) {
		if ( *it < 0x80 ) {
			ret += (char)*it;
		} else	{
			b2 = B2(*it);
			b3 = B3(*it);
			if ( *it < 0x800 ) { // 11bits
				 ch = 0xc0;
				 ch |= b2 << 2;
				 ch |= b3 >> 6;
				 ret += ch;

				 ch = 0x80;
				 ch |= b3 & 0x3f;
				 ret += ch;
			} else if ( *it < 0x10000 ) {
				ch = 0xe0;
				ch |= b2 >> 4;
				ret += ch;

				ch = 0x80;
				ch |= (b2 & 0x0f) << 2;
				ch |= b3 >> 6;
				ret += ch;

				ch = 0x80;
				ch |= b3 & 0x3f;
				ret += ch;
			} else if (  *it < 0x10FFFF  ) {
				b1 = B1(*it);
				ch = 0xf0;
				ch |= b1 >> 2;
				ret += ch;

				ch = 0x80;
				ch |= (b1&3) << 4;
				ch |= b2 >> 4;
				ret += ch;

				ch = 0x80;
				ch |= (b2&0xf) << 2;
				ch |= b3 >> 6;
				ret += ch;

				ch = 0x80;
				ch |= b3 & 0x3f;
				ret += ch;

			}
			assert( *it < 0x110000);
		}

		++ it;
	}
	return ret;
}


//int C0::translateChar() {
int Conv::doC0() {
	uint8_t byte = inByte(0);
	uint8_t byte1 = inByte(1);

	switch ( byte )
	{
		case ESC_CHAR:
			switch ( byte1 )
			{
				//呼び出し
				case 0x6e: //LS2
					ppGL = &pG2;
					return 2;

				case 0x6f: //LS3
					ppGL = &pG3;
					return 2;

				case 0x7e: //LS1R
					ppGR = &pG1;
					return 2;

				case 0x7d: //LS2R
					ppGR = &pG2;
					return 2;

				case 0x7c: //LS3R
					ppGR = &pG3;
					return 2;

				default:
					uint8_t F;
					int ret = 3;
					Translater **directee;
					// 1バイトGセット指示
					//uint8_t byte2 = inByte(2);
					if( byte1 >= 0x28 && byte1 <= 0x2b )
					{
						F = inByte(2);
						switch ( byte1 )
						{
							case 0x28:
								directee = &pG0;
								break;
							case 0x29:
								directee = &pG1;
								break;
							case 0x2a:
								directee = &pG2;
								break;
							case 0x2b:
								directee = &pG3;
								break;
						}
					}
					else if ( byte1 == 0x24 )
					{
						// 2バイトGセット指示
						switch ( inByte(2) )
						{
							case 0x29:
								F = inByte(3);
								ret = 4;
								directee = &pG1;
								break;
							case 0x2a:
								F = inByte(3);
								ret = 4;
								directee = &pG2;
								break;
							case 0x2b:
								F = inByte(3);
								ret = 4;
								directee = &pG3;
								break;
							default:
								F = inByte(2);
								ret = 3;
								directee = &pG0;
								break;
						}
					}
					else
					{
						cerr << "不明な指示 " << __FILE__ << " " << __LINE__ << endl;
						return 2; // 適当
					}

					switch ( F )
					{
						case 0x39: //JIS1
							*directee = &jis1;
							return ret;
						case 0x3a: //JIS2
							*directee = &no_op;
							return ret;
						case 0x4a: //英数
							*directee = &eisuu;
							return ret;
						case 0x30: //平仮名
							*directee = &hiragana;
							return ret;
						case 0x31: //片仮名
							*directee = &katakana_full_width;
							return ret;
						case 0x3b: //追加記号
							*directee = &tsuika_kigou;
							return ret;
						default:
							cerr << "未対応の指示<" << hex << (int)byte << ">"
									<< "<" << (int)byte1 <<  ">"
									<< "<" << (int)F <<  ">"
									<< endl;
							return ret; // 適当
					}
			} // switch ( byte1 )
			cerr << "未対応のC0<" << hex << (int)byte << ">"
					<< "<" << (int)byte1 <<  ">"
					<< endl;
			return 2; // 適当

//			break; //case ESC_CHAR:

		case LS0:
			ppGL = &pG0;
			return 1;

			break; //LS0

		case LS1:
			ppGL = &pG1;
			return 1;

			break; //LS1

		case SS2: // SSが来てGRとかマルチバイトが来たらどうなんのだろう？
						++pIn;
						return pG2->translateChar();
			break; //SS2

		case SS3:
						++pIn;
						return pG3->translateChar();
						break; //SS3

		case APR:
			pStrOut->push_back( '\n' );
			return 1;

		case SP:
			pStrOut->push_back( SPACE[f_h]);
			return 1;

		default:
#if 1
			//無視してみるテスト
			cerr << "未対応のC0<" << hex << (int)byte <<  ">" << endl;
			return 1;

#else
			// 他のコントロールキャラはそんまま
			pStrOut->push_back(byte);
			return 1;
#endif
			break;

	} // switch ( byte )

	return 0;
}

int Conv::doC1() {
	uint8_t byte = inByte(0);

	switch ( byte )
	{
	 case MSZ:
		 f_h = 1;
		 return 1;

	 case NSZ:
		 f_h = 0;
		 return 1;

	 case CSI:
		 return 1;
	}
	cerr << "未対応のC1<" << hex << (int)byte <<  ">" << endl;

	return 0;
}


int Conv::continued( const uint8_t *pInput, int len, StrType* pOut ) {
	pIn = pInput;
	pStrOut = pOut;

	const uint8_t* pStart = pInput;
	const uint8_t *pEnd = pIn + len;

//		enum Region { C0 = 0, GL = 1, C1 = 2, GR = 3 } region;
	int ret = 0;


	while ( pIn < pEnd ) {
//		unit8_t *p;
//		p = pIn;

		if ( *pIn < 0x21 ) {
			ret = doC0();
		} else if ( *pIn < 0x80 ) {
			ret = (*ppGL)->translateChar();
		} else if ( *pIn < 0xa0 ) {
			ret = doC1();
		} else {
			ret = (*ppGR)->translateChar();
		}

		if ( ret < 1 ) {
			return pStart - pIn;
		}

		pIn += ret;

	}
	return 0;
}


int Eisuu::translateChar() {	 //QByteArray& baOut, const uint8_t *pIn ) {
		uint16_t c = (inByte(0) & 0x7f) ;
		pStrOut->push_back(c);
		return 1;
}
/*
int EisuuZenkaku() {
}
*/

static const uint16_t katakana_nokori[] = { 0x30fd, 0x30fe, 0x30fc/*ー*/ , 0x3002, 0x300c, 0x300d, 0x3001, 0x30fb, /*DEL*/0x2588 };
// ctrl DELETE = 0x007f
// SYMBOL FOR DELETE = 0x2421
// U+2588 FULL BLOCK
// U+2B1B BLACK LARGE SQUARE


int Katakana_full_width::translateChar() {
		static const uint16_t head = 0x30a1;
		uint16_t code;
		uint8_t ch;

		ch = inByte(0) & 0x7f;
/*
		if ( ch == 0x20 || ch == 0x7f ) {
			pStrOut->push_back(' ');
			return 1;
		} else
*/ 	if ( ch <= 0x76 ) { // ヶ まで
			code = head + ( ch - 0x21 );
		} else {
			code = katakana_nokori[ch - 0x77];
		}

		pStrOut->push_back( code );
		return 1;
}

static const uint16_t hiragana_nokori[] = { 0x309d, 0x309e, 0x30fc, 0x3002, 0x300c, 0x300d, 0x3001, 0x30fb, 0x2588 };

int Hiragana::translateChar() {
		static const uint16_t head = 0x3041;
		uint8_t ch;
		uint16_t code;

		ch = inByte(0) & 0x7f;
/*		if ( ch == 0x20 || ch == 0x7f ) {
			pStrOut->push_back(' ');
			return 1;
		} else
*/
		if ( ch <= 0x76 ) { // ゔゕゖ　もいれちゃってるけど
			code = head + ( ch - 0x21 );
		} else {
			code = hiragana_nokori[ch - 0x77];
		}

		pStrOut->push_back( code );

		return 1;
}

int No_op::translateChar() {
		return 1;
}


int Jis1::translateChar() {
//		static const uint16_t braceF = 0xff1c, braceT = 0xff1e;
		static const uint16_t braceF = '[', braceT = ']';

		char mb[2] , *pib = mb;
		mb[0] = ( inByte(0) | 0x80 );
		mb[1] = ( inByte(1) | 0x80 );

/*
  if 0x2121 SPACE[f_h]

 */

		uint16_t uch[2], *puch = uch;
		size_t ret, ibl = 2, obl = 4;

		ret = iconv( cd, &pib, &ibl, (char**)&puch, &obl );

		if ( 0 == ibl ) {
			int chsize = (4 - obl ) / 2;
			for ( int i = 0; i < chsize; ++ i ) {
				pStrOut->push_back( uch[i] );
			}
		} else {
			iconv( cd, NULL, NULL, NULL, NULL );

			uint8_t n4;
			pStrOut->push_back( braceF );
//			uint8_t ib = inByte(0);

			n4 = inByte(0) >> 4;
			uch[0] = n4 > 9 ? 'a' + n4 - 10 : '0' + n4;
			pStrOut->push_back( uch[0] );

			n4 = inByte(0) & 0x0f;
			uch[0] = n4 > 9 ? 'a' + n4 - 10: '0' + n4;
			pStrOut->push_back( uch[0] );

			pStrOut->push_back( braceT );


			pStrOut->push_back( braceF );

			n4 = inByte(1) >> 4;
			uch[0] = n4 > 9 ? 'a' + n4 - 10: '0' + n4;
			pStrOut->push_back( uch[0] );

			n4 = inByte(1) & 0x0f;
			uch[0] = n4 > 9 ? 'a' + n4 - 10: '0' + n4;
			pStrOut->push_back( uch[0] );

			pStrOut->push_back( braceT );
		}
		return 2;
}

int Tsuika_kigou::translateChar() {
		int b0 = inByte(0);
		int b1 = inByte(1);
		b0 &= 0x7f;
		b1 &= 0x7f;
		int idx_ku = b0 - 0x75;
		int idx_ten = b1 - 0x21;

		if ( idx_ten < 0 )
		{
			goto ERROR;
		}

		if ( b0 >= 0x75 && b0 <= 0x7e )
		{
			if ( idx_ten < TUIKA_KIGOU_LEN[idx_ku] )
			{
				pStrOut->push_back( TUIKA_KIGOU[idx_ku][idx_ten] );
				return 2;
			}
		}
		ERROR:
		cerr << hex << "範囲外の追加記号<" << b0 << ">"
				<< "<" << b1  << ">"
				<< endl;
		return 2;
}

/////////////////////////////////////////////
void aribDay( const uint8_t* pAribDate /*2 bytes*/, struct tm *pOut ) {
	unsigned uiMjd; //MJD local time 日本時間1858年11月17日00:00からの日数？
	double mjd, dy_, dm_;
	int y_, m_, d, k, y, m;

//	unsigned uiMjd1900 = 15020; // 1900/1/1
	static const unsigned uiMjd2000_01_01 = 51544; // 2000/1/1
//	unsigned uiMjd2010_01_01 = 55197; // 2010/1/1
//	unisgned uiMjd65535; // 2038/4/22


	uiMjd = ( pAribDate[0] << 8 ) | pAribDate[1]; // MJDの下位16bitしか無い。有効期限2038年のとある日
#if 1
	unsigned mjdTEKITOU = uiMjd2000_01_01;
	if ( uiMjd < mjdTEKITOU )
	{
		uiMjd |= 0x10000;
	}
#endif

	mjd = uiMjd;
	dy_ = floor( (mjd - 15078.2) / 365.25 );
	dm_ = floor( (mjd - 14956.1 - floor( dy_ * 365.25 ) ) / 30.6001 );
	d = uiMjd - 14956 - floor( dy_ * 365.25) - floor( dm_ * 30.6001 );

	y_ = dy_;
	m_ = dm_;

	if ( m_ > 13 )
	{
		k = 1;
	}
	else
	{
		k = 0;
	}

	y = y_ + k;
	m = m_ - 2 - k * 12;

	pOut->tm_year = y;
	pOut->tm_mon = m;
	pOut->tm_mday = d;
	pOut->tm_wday = (uiMjd +4) % 7;
};

string& hexstr( const uint8_t* p, int len, string& s )
{
	s = "";
	for ( int i = 0; i < len; ++ i)
	{
		char ch,cl;
		ch = p[i] >> 4;
		cl = p[i] & 0x0f;
		s += "<";
		s += ch >9 ? 'A' + (ch - 10) : '0' + ch;
		s += cl >9 ? 'A' + (cl - 10) : '0' + cl;
		s += ">";
	}
	return s;
};
