/**************************************************************************
**   yapoo 2010
**************************************************************************/
#ifndef ARIBSTRING_H
#define ARIBSTRING_H

#include <iostream>
#include <stdlib.h>
#include <string>
#include <stdint.h>
#include <vector>
#include <iconv.h>
#include <endian.h>
#include <time.h>
#include <cassert>
#if __BYTE_ORDER == __LITTLE_ENDIAN
	static const char iconvTo[] = "UTF-16LE";
//	static const char iconvTo = "UTF-16LE//TRANSLIT"
//	static const char iconvTo = "UTF-16LE//IGNORE"
#else
	static const char iconvTo[] = "UTF-16BE";
#endif

using namespace std;

typedef char32_t CharType;
typedef vector<char32_t> StrType;
//typedef wstring StrType;

//extern const char16_t k85[],k86[],k90[],k91[],k92[],k93[],k94[];
extern char16_t* const TUIKA_KIGOU[];
extern int const TUIKA_KIGOU_LEN[];

string& hexstr( const uint8_t* p, int len, string& s );

class Translater {
	private:
	const uint8_t* &pIn;
//	Translater(){};

	protected:
	StrType* &pStrOut;

	// const uint8_t* inPos( int pos ) { return &pIn[pos]; };
	uint8_t inByte( int i ) { return pIn[i]; }

	public:
	Translater( StrType* &pStr, const uint8_t* &p ) : pIn( p ), pStrOut( pStr ) {

	};
	virtual int translateChar() = 0;// QByteArray& baOut, const uint8_t *pIn ) = 0;
};

class Eisuu : public Translater {
	public:
	Eisuu(StrType* &pStr, const uint8_t* &p):Translater( pStr, p ) {};
	int translateChar() ;
};

class Hiragana : public Translater {
	public:
	Hiragana(StrType* &pStr, const uint8_t* &p):Translater( pStr, p ) {};
	int translateChar() ;
};

class Katakana_full_width : public Translater {
	public:
	Katakana_full_width(StrType* &pStr, const uint8_t* &p):Translater( pStr, p ) {};
	int translateChar() ;
};

class Jis1 : public Translater {
protected:
	iconv_t cd;
	const char* iconvFrom;
	public:
	Jis1(StrType* &pStr, const uint8_t* &p): Translater( pStr, p ),
	iconvFrom("EUC-JP")
//	iconvFrom( "EUC-JISX0213" )
	{
		cd = iconv_open( iconvTo, iconvFrom );
		if ( (iconv_t)-1 == cd ) {
			cerr << "iconv error" << __FILE__ << ":" << __LINE__ << endl;
			exit( -1 );
		}

	}

	~Jis1() {
		if ( (iconv_t)-1 != cd ) {
			iconv_close( cd );
		}
	}

	int translateChar() ;
};

class Tsuika_kigou : public Translater {
public:
	Tsuika_kigou (StrType* &pStr, const uint8_t* &p) : Translater ( pStr, p ) {}
	int translateChar() ;

};

class No_op : public Translater {
	public:
	No_op(StrType* &pStr, const uint8_t* &p):Translater( pStr, p ) {};
	int translateChar() ;
};

/*
class C0 : public Translater {
	public:
	C0(StrType* &pstrOut, const uint8_t* &p):Translater( pstrOut, p ) {};
	int translateChar() ;
};


class C1 : public Translater {
	public:
	C1(StrType* &pstrOut, const uint8_t* &p):Translater( pstrOut, p ) {};
	int translateChar() { return 0; };
};
*/


class Conv {
//	friend class L0;
//	enum { SingleByte, MulitByte } charStatus;

	const uint8_t *pIn;
	StrType *pStrOut;

	Eisuu eisuu;
	Jis1 jis1;
	Hiragana hiragana;
	Katakana_full_width katakana_full_width;
	Tsuika_kigou tsuika_kigou;
	No_op no_op;
	int f_h; //full \ half width
	//	C0 c0;
//	C1 c1;

//	QByteArray outBuff;
	Translater **ppGL, **ppGR;
	Translater *pG0, *pG1, *pG2, *pG3;
//	Translater *pC0, *pC1;


//	QByteArray* pbaOut;



	uint8_t inByte( int i ) { return pIn[i]; }
	int doC0();
	int doC1();

public:
	Conv()
	: /* strOut( "" ), */
	eisuu( pStrOut, pIn ),
	jis1( pStrOut, pIn ),
	hiragana( pStrOut, pIn ),
	katakana_full_width( pStrOut, pIn ),
	tsuika_kigou( pStrOut, pIn ),
	no_op( pStrOut, pIn )
//	, c0( pstrOutOut, pIn )
//	, c1( pstrOutOut, pIn )
	 {
		resetState();
	 };



	void resetState() {
		pG0 = &jis1;
		pG1 = &eisuu;
		pG2 = &hiragana;
		pG3 = &katakana_full_width;
		f_h = 0;

		ppGL = &pG0;
		ppGR = &pG2;

//		pStrOut->clear();
	};


	int continued( const uint8_t *pInput, int len, StrType* pOut );
	int convert( const uint8_t *pInput, int len, StrType* pOut ) {
		resetState();
		pOut->clear();
		return continued( pInput, len, pOut );
	};

	string utf8();
};

void aribDay( const uint8_t* pAribDate /*2 bytes*/, struct tm *pOut );
void aribTime( const uint8_t* pAribDate /*3 bytes*/, struct tm *pOut );
void aribDayTime( const uint8_t* pAribDate /*5 bytes*/, struct tm *pOut );
inline int BCD( uint8_t bcd )
{
	int ret;
	ret = (bcd >> 4) * 10;
	ret += (bcd & 0x0f);
	assert( ret >= 0 && ret < 100 );
	return ret;
}

inline void aribTime( const uint8_t* pAribDate /*3 bytes*/, struct tm *pOut )
{
	pOut->tm_hour = BCD( pAribDate[0] );
	pOut->tm_min = BCD( pAribDate[1] );
	pOut->tm_sec = BCD( pAribDate[2] );
};

inline void aribDayTime( const uint8_t* pAribDate /*5 bytes*/, struct tm *pOut )
{
	aribDay( pAribDate, pOut );
	aribTime( &pAribDate[2], pOut );
};


#endif // ARIBSTRING_H
