/**************************************************************************
**   yapoo 2010
**************************************************************************/
#include "mainwindow.h"
#include "backendthread.h"
#include "ui_mainwindow.h"
#include "psisection.h"
//#include <termios.h>
#include <signal.h>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
	inputProcess = 0;
	inputBuffer = 0;

	playerProcess = 0;
	playerThread = 0;

	siThread = 0;

//	fd_in = -1;
//	fd_out = -1;
	recpt1_pid = -1;

//	qRegisterMetaType<QProcess::ProcessState>("QProcess::ProcessState");
	qRegisterMetaType<Eit_PfInfoLocker*>("Eit_PfInfoLocker*");
	connect(ui->eit_pf12View, SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)),
				this, SLOT(onEit_pfViewCurrentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)));
	connect(ui->eit_pf12View, SIGNAL(itemDoubleClicked(QTreeWidgetItem*,int)),
				this, SLOT(onProgramSelected(QTreeWidgetItem*,int)));
}

MainWindow::~MainWindow()
{
	cleanupTasks();
	::unlink( fifoname_in.toLocal8Bit() );
	::unlink( fifoname_out.toLocal8Bit() );
	delete ui;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

int MainWindow::setupConfig( const char *filename )

{
	pid_t pid;
	int ret;
	bool fRet;

	player_name = "mplayer";
	recpt1_name = "recpt1";
	fifo_dir = "/tmp";

	player_arg = (QStringList()
							<< "-slave" << "-input" << "nodefault-bindings:conf=/dev/null"
							<< "-vf" << "pp=lb"
							<< "-cache" << "3000"
							<< "-cache-min" << "70"
							<< "-quiet"
							<< "-autosync" << "0"
							);
	recpt1_arg = (QStringList() << "--b25" << "--strip");

	if ( filename )
	{
		config_filename = filename;
	}
	else
	{
		config_filename = QDir::homePath() + "/qmpt.config";
	}

	cerr << "config file: " << (const char*)config_filename.toLocal8Bit() << endl;

	QFile file(config_filename);
	fRet = file.open(QIODevice::ReadOnly);
	if ( fRet == true )
	{
		readConfig( file );
	}
	else
	{
		cerr << "file not found" << endl;
	}

	pid = ::getpid();
	fifoname_in = QString( "%1/fifo_in_qmpt%2").arg(fifo_dir).arg( pid );
	ret = ::mkfifo( fifoname_in.toLocal8Bit(), S_IRUSR | S_IWUSR );
	if ( ret )
	{
		QMessageBox::critical( this, QString::fromLocal8Bit("FIFOの作成に失敗しました"), fifoname_in );
		return -1;
	}
	fifoname_out = QString( "%1/fifo_out_qmpt%2").arg(fifo_dir).arg( pid );
	player_arg << fifoname_out;

	return 0;
}


void MainWindow::readConfig( QFile& file ) {
	int ret;
	QString stringLine;
	QString tmp;
	QStringList stringList;
	QPushButton *pushButton;

	const char button[] = "channelButton";
	const char player[] = "mplayer=";
	const char recpt1[] = "recpt1=";
	const char player_arg[] = "mplayer_arg=";
	const char recpt1_arg[] = "recpt1_arg=";
	const char fifo_dir[] = "fifo_dir=";

	do{
		ret = readLine( file, stringLine );
		if ( stringLine.startsWith(button,Qt::CaseInsensitive))
		{
			tmp = stringLine.simplified();
			stringList = tmp.split("=");
			pushButton = this->findChild<QPushButton *>( stringList.at(0) );
			if (button == 0)
			{
				continue;
			}
			pushButton->setText( stringList.value(1) );
			pushButton->setProperty( "channel", QVariant(stringList.value(2)));
			pushButton->setProperty( "default_program_id", QVariant(stringList.value(3).toInt()));
		}
		else if ( stringLine.startsWith(player,Qt::CaseInsensitive))
		{
			this->player_name = stringLine.mid( ::strlen(player) ).simplified();
		}
		else if ( stringLine.startsWith(recpt1,Qt::CaseInsensitive))
		{
			this->recpt1_name = stringLine.mid( ::strlen(recpt1) ).simplified();
		}
		else if ( stringLine.startsWith(fifo_dir,Qt::CaseInsensitive))
		{
			this->fifo_dir = stringLine.mid( ::strlen(fifo_dir) ).simplified();
		}
		else if ( stringLine.startsWith(player_arg,Qt::CaseInsensitive))
		{
			tmp = stringLine.mid( ::strlen(player_arg) ).simplified();
			this->player_arg = tmp.split(" ", QString::SkipEmptyParts);
		}
		else if ( stringLine.startsWith(recpt1_arg,Qt::CaseInsensitive))
		{
			tmp = stringLine.mid( ::strlen(recpt1_arg) ).simplified();
			this->recpt1_arg = tmp.split(" ", QString::SkipEmptyParts);
		}
	} while( !ret );
}

int MainWindow::readLine( QFile& file, QString& stringLine )
{
	qint64 iRet;
	int ret = 0;
	QByteArray baLine;
	char line[1024];

//	baLine.clear();
	do {
		iRet = file.readLine(line, sizeof(line));
		if (iRet <= 0 )
		{
			ret = -1;
			break;
		}
		if ( iRet >= 2 && line[iRet - 2] == '\\' )
		{
			baLine.append(line,iRet - 2);
			continue;
		}
		baLine.append(line, iRet - 1);
		break;
	} while ( true );

	stringLine = stringLine.fromLocal8Bit( baLine.constData() );
	return ret;
}


void MainWindow::cleanupTasks()
{
	cerr << "cleanupTasks()" << endl;

	int ret;

	if ( playerProcess )
	{
		stopPlayer();
	}

	if ( inputProcess )
	{
		disconnect( inputProcess, SIGNAL(finished( int, QProcess::ExitStatus )),
					this, SLOT(onInputProcessFinished( int, QProcess::ExitStatus )));
	recpt1_pid = -1;
	inputProcess->closeWriteChannel();
	inputProcess->closeReadChannel(QProcess::StandardOutput);
	inputProcess->closeReadChannel(QProcess::StandardError);

	inputProcess->close();
	ret = inputProcess->waitForFinished(10000);
	cerr << "input processes finished: "<< ret << endl;
	delete inputProcess;
	inputProcess = 0;

	}

	if ( inputBuffer )
	inputBuffer->exitFlag = 1;
	if( siThread )
	siThread->exitFlag = 1;
	if (playerThread )
	playerThread->exitFlag= 1;

	if ( playerThread )
	{
		if ( !playerThread->wait(10000) )
		{
			cerr << "playerThread->wait() timed out" << endl;
		}
		delete playerThread;
		playerThread = 0;
	}
	if ( siThread )
	{
		if ( !siThread->wait(10000) )
		{
			cerr << "siThread->wait() timed out" << endl;
		}
		delete siThread;
		siThread = 0;
	}
	if ( inputBuffer )
	{
		if (!inputBuffer->wait(10000)) {
			cerr << "inputBuffer->wait() timed out" << endl;
		}
		delete inputBuffer;
		inputBuffer = 0;
	}
	ui->eit_pf12View->clear();
	ui->plainTextEdit->clear();
	infoLocker.lock().clear();
	infoLocker.unlock();
}

void MainWindow::onEit_PfChanged()
{
	ui->eit_pf12View->clear();
	QString format;
	QChar fill = '0';
	QTreeWidgetItem *item, *item2;
	Infos::const_iterator it;
	Infos &infos = infoLocker.lock();

	it = infos.begin();
#if 0
	if ( it != infos.end())
	{
		if ( playerProcess )
		{
			QString str = QString("set_property switch_program %1\n")
						  .arg( it->second->service_id );
			playerProcess->write( str.toAscii() );
			qDebug() << str;
		}
	}
#endif
	for(;
		it != infos.end();
		++ it )
	{
		spEit_pfInfo pInfo = it->second;
		if (pInfo->version == -1 )
		{
			continue;
		}
		format = QString( "service:%1 event:%2 ver:%3" )
				.arg(pInfo->service_id)
				.arg(pInfo->event_id)
				.arg(pInfo->version);
		item = new QTreeWidgetItem( ui->eit_pf12View );
		item->setText(0,format);

		item->setData(1,128,QVariant(pInfo->service_id));

		item2 = new QTreeWidgetItem( item );
		item2->setText(0,pInfo->title);

		item2 = new QTreeWidgetItem( item );
		format = QString::fromUtf16((const ushort*)u"%1〜%2")
				.arg(pInfo->start.toString())
				.arg(pInfo->end.toString());
		item2->setText(0,format);

		item2 = new QTreeWidgetItem( item );
		item2->setText(0,pInfo->description);

		item->setExpanded(true);
/*
		vector<u16pair>::iterator it2;
		for ( it2 = info.eses.begin(); it2 != info.eses.end(); ++ it2 )
		{

			format = QString( "Type: 0x%1 PID=0x%2" )
							.arg(it2->first,2,16,fill)
							.arg(it2->second,4,16,fill);
			item2 = new QTreeWidgetItem( item );
			item2->setText(0,format);
		}
		*/
	}
	infoLocker.unlock();

}

void MainWindow::startPlayer( int program_id )
{
	QStringList arg = this->player_arg;
	int ret;
	player_error_count = 0;
	error_watch_timer.setInterval( 2000 );
	connect( &error_watch_timer, SIGNAL(timeout()), this, SLOT( onErrorWatchTimerTimedOut() ));
	error_watch_timer.start();

	current_prog = program_id;
	::unlink( fifoname_out.toLocal8Bit() );
	ret = ::mkfifo( fifoname_out.toLocal8Bit(), S_IRUSR | S_IWUSR );
	if ( ret )
	{
		QMessageBox::critical( this, QString::fromLocal8Bit("FIFOの作成に失敗しました"), fifoname_out );
		return;
	}
	if ( program_id > 0 )
	{
		arg << "-tsprog" << QString::number(program_id);
	}

	playerProcess = new QProcess();
	playerProcess->setProcessChannelMode(QProcess::SeparateChannels);
	playerProcess->setReadChannel(QProcess::StandardError);
	connect(playerProcess, SIGNAL(readyRead()),
				this, SLOT(onPlayerStderrRecieved()));

	qDebug() << player_name;
	qDebug() << arg;

	playerProcess->setEnvironment(QProcess::systemEnvironment());

	playerProcess->setStandardOutputFile( "/dev/null");
//	playerProcess->setStandardErrorFile( "/dev/tty");

	playerProcess->start( player_name, arg, QIODevice::ReadWrite | QIODevice::Unbuffered );
	connect( playerProcess, SIGNAL(finished(int, QProcess::ExitStatus)),
			 this, SLOT(onPlayerProcessFinished(int,QProcess::ExitStatus)));

	playerThread = new PlayerThread( inputBuffer, playerProcess, fifoname_out.toLocal8Bit() );
	playerThread->start();
	playerProcess->waitForStarted();
}


void MainWindow::stopPlayer()
{
	if ( !playerProcess )
	{
//		cerr << "playerProcess = 0" << endl;
		return;
	}
	cerr << "waiting player quits" << endl;
	disconnect( playerProcess, SIGNAL(finished( int, QProcess::ExitStatus )),
				this, SLOT(onPlayerProcessFinished( int, QProcess::ExitStatus )));

	error_watch_timer.stop();
	disconnect( &error_watch_timer, SIGNAL(timeout()), this, SLOT( onErrorWatchTimerTimedOut() ));

	if (!playerProcess->waitForFinished(100))
	{
		playerProcess->write("quit\n");
		cerr << "send quit" << endl;
	}
/*
	playerProcess->closeWriteChannel();
	playerProcess->closeReadChannel(QProcess::StandardOutput);
	playerProcess->closeReadChannel(QProcess::StandardError);
*/
	if (!playerProcess->waitForFinished(5000))
	{
		cerr << "killing player processes" << endl;
		playerProcess->close();
		if (!playerProcess->waitForFinished(1000)) {
			playerProcess->close();
			if (!playerProcess->waitForFinished(2000)) {
				playerProcess->close();
				if (!playerProcess->waitForFinished(5000)) {
					playerProcess->kill();

					QProcess killall;
					QStringList arg;
					cerr << "execute killall -KILL mplayer" << endl;
					arg << "-KILL" << "mplayer";
					killall.execute( "killall", arg );
					sleep(1);
				}
			}
		}
	}
	else
	{
		cerr << "player processes finished" << endl;
	}

/*
	if ( fd_out > -1 )
	{
	//	::close(fd_out);
		fd_out = -1;
	}
*/
	disconnect(playerProcess,SIGNAL(readyRead()),this,SLOT(onPlayerStderrRecieved()));
	delete playerProcess;
	playerProcess = 0;
//	sleep(2);
}

void MainWindow::startInput( const char *gr )
{
		QStringList env = QProcess::systemEnvironment();

		QStringList recpt1Arg = recpt1_arg;
		recpt1Arg << gr << "-" << fifoname_in;
		inputProcess = new QProcess();
		inputProcess->setProcessChannelMode(QProcess::ForwardedChannels);
		//	inputProcess->setStandardErrorFile( "/dev/tty" );
		inputProcess->setEnvironment(env);

		inputProcess->start( recpt1_name, recpt1Arg ); //,   QIODevice::ReadOnly | QIODevice::Unbuffered );
		connect( inputProcess, SIGNAL(finished( int, QProcess::ExitStatus )),
					this, SLOT(onInputProcessFinished( int, QProcess::ExitStatus )));
		inputProcess->waitForStarted();
	//	inputProcess->waitForReadyRead(5000);
		recpt1_pid = inputProcess->pid();

		inputBuffer = new InputBuffer( fifoname_in.toLocal8Bit() );
		inputBuffer->start();

		siThread = new SiThread( inputBuffer, this );
		siThread->start();
}

void MainWindow::onInputProcessFinished( int, QProcess::ExitStatus )
{
	cleanupTasks();
}

void MainWindow::onPlayerProcessFinished( int, QProcess::ExitStatus )
{
	stopPlayer();
}


void MainWindow::onChannelButtonClicked()
{
	QVariant var,var2;
	QString channel;
//	QProcess pt1ctl;
//	volatile int ts_id = -1;
//	QString pt1ctl_name = "recpt1ctl";
	QList<QPushButton*> buttons;
	QList<QPushButton*>::iterator it;
	int program_id = 0;

	var = sender()->property( "channel" );
	if (!var.isValid())
	{
		return;
	}
	var2 = sender()->property( "default_program_id" );
	if ( var2.isValid() )
	{
		program_id = var2.toInt();
	}

	buttons = sender()->parent()->findChildren<QPushButton*>();
	for( it = buttons.begin();
		it != buttons.end();
		++ it )
	{
		(*it)->blockSignals( true );
	}
	cleanupTasks();

	channel = var.toString();

	startInput( channel.toAscii() );
	startPlayer( program_id );

	for( it = buttons.begin();
		it != buttons.end();
		++ it )
	{
		(*it)->blockSignals( false );
	}
}

void MainWindow::onEit_pfViewCurrentItemChanged(QTreeWidgetItem* item, QTreeWidgetItem*)
{
	if ( item )
	{
		ui->plainTextEdit->setPlainText(item->text(0));
	}
}

void MainWindow::onProgramSelected(QTreeWidgetItem* pItem, int)
{

	QVariant data = pItem->data(1,128);
	if ( data.isNull() )
	{
		return;
	}

	if ( !playerProcess )
	{
		return;
	}

	QByteArray command( "set_property switch_program ");
	command += data.toString();
	command += "\n";

	qDebug() << command;

	playerProcess->write( command );

}

void MainWindow::onPlayerStderrRecieved()
{
	int count;
	QString str = QString(playerProcess->readAll());
	count = str.count( "resync" );
	ui->plainTextEdit_2->appendPlainText( str );
	player_error_count += count;
	if ( player_error_count > 10 )
	{
		stopPlayer();
		::sleep(1);
		startPlayer( current_prog );
	}
}

void MainWindow::onPlayerCommandRecieved()
{
	if ( !playerProcess )
	{
		return;
	}
	QString s = ui->playerCommandEdit->text();
	s += "\n";
	playerProcess->write( s.toAscii() );
}

