/**************************************************************************
**   yapoo 2010
**************************************************************************/
#include "psisection.h"
#include "descriptor.h"
#include "mainwindow.h"


SiThread::SiThread( InputBuffer *p, MainWindow *mainwindow )
{
	inputBuffer = p;
	ts_id = -2;
	reader = new InputBuffer::Reader( p );
	eitOp12 = new Eit_pfOperator( mainwindow );
	eitOp27 = new Eit_pfOperator( mainwindow );
}


SiThread::~SiThread()
{
	delete reader;
	delete eitOp12;
	delete eitOp27;
}


void SiThread::run()
{
	cerr << "SiThread start: thread_id:" <<  currentThreadId () << endl;
	start:
/*
	int ret;
	ret = waitForTsChanged(2);
	if ( ret )
	{
		exitFlag = 0;
		return;
	}
*/
	exitFlag = 0;

	do
	{
		doRead();
		msleep(1000/100);
	} while ( !exitFlag );

	if ( exitFlag == 2 )
	{
		goto start;
	}
	cerr << "SiThread end: thread_id:" <<  currentThreadId () << endl;
}

int SiThread::waitForTsChanged( time_t wait_seconds )
{
	size_t ret;
	char *ts;
	TsHeadParser thp;
	time_t start_time = ::time( NULL );
	uint16_t new_id;
	do{
		ret = reader->read( &ts, 188 );
		if ( !ret )
		{
			msleep(100);
			continue;
		}
		thp.setPos( ts );
		assert( thp.syncOK() );
		if ( thp.pid() != 0 || !thp.isStart() )
		{
			continue;
		}
		new_id = (ts[3]<<8)|ts[4];
		if (ts_id != new_id )
		{
			ts_id = new_id;
			return 0;
		}
	} while ( time(NULL) - start_time  < wait_seconds );
	return -1;
}

void SiThread::doRead()
{
	int read_size;
	const uint8_t* p;

	read_size = reader->read( (char**)&p, 188*10000 );
	if ( read_size == 0 )
	{
		msleep( 1000/120 );
		return;
	}

	TsHeadParser tsh;
	for ( int i = 0; i < read_size; i += 188 )
	{
		tsh.setPos( &p[i] );
assert ( tsh.syncOK() );

		switch ( tsh.pid() )
		{
			case 0x12:
				eitOp12->putPacket( &p[i]);
				break;

 /*
			case 0x27:
				eitOp27->putPacket( &p[i]);
				break;
*/
		}
	}
}


Eit_pfOperator::Eit_pfOperator( MainWindow* p )
{
//		info.version = -1;
//		pController = p;
		tsid = -1;
		infoLocker = &p->infoLocker;
		mainWindow = p;
/*
	bool fRet;
	fRet = connect( this, SIGNAL(eit_pf_changed(Eit_PfInfoLocker*)),
	 p, SLOT(onEit_PfChanged(Eit_PfInfoLocker*)),
	 Qt::QueuedConnection);
*/
	connect( this, SIGNAL(eit_pf_changed()),
	 p, SLOT(onEit_PfChanged()),
	 Qt::QueuedConnection);

//	cout << fRet << endl;

	reader.setBuffer( &buff );
}

Eit_pfOperator::~Eit_pfOperator()
{
	disconnect( this, SIGNAL(eit_pf_changed()),
		 mainWindow, SLOT(onEit_PfChanged()) );
}


void Eit_pfOperator::putPacket( const uint8_t* TsPacket )
{
	int ret;
	EitParser parser;

	ret = reader.push( TsPacket );

	while (ret == TsPacketReader::COMPLETE)
	{
		parser.setPos( buff.buff );
/*
		uint16_t tsid = parser.transport_stream_id();
		if ( this->tsid != tsid )
		{
			this->tsid = tsid;
			infoLocker->lock().clear();
			infoLocker->unlock();
			//emit
		}
*/
		if ( parser.table_id() != 0x4e )
		{
			reader.clearBuffer();
			ret = reader.next();
			continue;
		}

		if ( parser.section_number() != 0 )
		{
			reader.clearBuffer();
			ret = reader.next();
			continue;
		}

		if ( reader.checkCrc() )
		{
			cerr << "CRC error: " << __FILE__ << " " << __LINE__ << endl;
			reader.clearBuffer();
			ret = reader.next();
			continue;
		}

		Infos::iterator itEit;
//		InfoPair oldEit;
		spEit_pfInfo info;
		uint16_t service_id;

		service_id = parser.service_id();
		Infos& infos = infoLocker->lock();
		itEit = infos.find( service_id );
		if ( itEit == infos.end() )
		{
			info = spEit_pfInfo( new Eit_pfInfo );

			info->version = -1;
			infos[service_id] = info;
		}
		else
		{
			info = itEit->second;
		}

		if ( info->version == parser.version_number() )
		{
			reader.clearBuffer();
			ret = reader.next();
			infoLocker->unlock();
			continue;
		}

		Conv conv;
		StrType converted;

		info->version = parser.version_number();
		info->service_id = service_id;

		// ループは必ず１
		QString str;
		info->event_id = parser.i.event_id();
		const uint8_t *p = parser.i.start_time();
		struct tm tmTm;
		::aribDayTime( p, &tmTm );
		info->start.setHMS( tmTm.tm_hour, tmTm.tm_min, tmTm.tm_sec );
		int duration = BCD(parser.i.duration()[0]) * 60 * 60
				+ BCD( parser.i.duration()[1]) * 60
				+ BCD( parser.i.duration()[2]);
		info->end = info->start.addSecs( duration);
		info->descriptor.assign( parser.i.j.pHead, parser.i.j.pLimit);
//		int i = 0;

		while( parser.i.j.pHead < parser.i.j.pLimit )
		{
			unsigned tag;
			tag = parser.i.j.pHead[0];

			if ( tag == 0x4d )
			{
				ShortEventDescriptor sed(&parser.i.j.pHead[0]);
				conv.convert( sed.event_name_char(), sed.event_name_length(), &converted );

//	string cstr;
//	cerr << hexstr(sed.event_name_char(), sed.event_name_length(), cstr  ) << endl;
				info->title = QString::fromUcs4( (const uint*)&*(converted.begin()), converted.size() );

				conv.convert( sed.text_char(), sed.text_length(), &converted );
				info->description = QString::fromUcs4((const uint*)&*(converted.begin()), converted.size() );
//cerr << hexstr(sed.text_char(), sed.text_length(), cstr  ) << endl;
//cerr << conv.utf8() << endl;
			}

			++ parser.i.j;
		}

		reader.clearBuffer();
		ret = reader.next();
		infoLocker->unlock();
		emit eit_pf_changed();
//		pController->onEit_pfChanged( this );
	}

	if ( ret & TsPacketReader::ERR_ANY )
	{
		cerr << "packet read error: 0x" << hex << ret << dec << " " << __FILE__<< " " << __LINE__ << endl;
		reader.clearBuffer();
	}

}
