/**************************************************************************
**   yapoo 2010
**************************************************************************/
#ifndef PSISECTION_H
#define PSISECTION_H

#include <QtCore>

#include <stdint.h>
#include <memory>
#include <map>
#include <vector>

class MainWindow;
//#include "backendthread.h"
#include "TsPacketReader.h"
#include "aribstring.h"

using namespace std;

//class BackendThread;

#define UM16(x) ( ( (pHead[x]<<8)|pHead[x+1] )  )
#define M16(x,y) ( UM16(x) & ( 0x0FFFF>>(16-y) ) )

class PsiCommon {
  protected:
	const uint8_t *pHead;

  public:
	uint8_t table_id() { return *pHead; }
	bool section_syntax_indicator() { return (pHead[1] & 0x80); }

	uint8_t section_length() { return M16(1,12); }
	uint16_t transport_stream_id() { return M16(3,16); }

	uint8_t version_number() { return (rawVer() >> 1); }

	uint8_t rawCni() { return (pHead[5] & 0x01); }
	bool current_next_indicator() { return rawCni(); }
	uint8_t section_number() { return pHead[6]; }
	uint8_t last_section_number() { return pHead[7]; }


	uint8_t rawVer() { return (pHead[5] & 0x3e); }
};

class EitParser : public PsiCommon {
  public:

	void setPos( const uint8_t *pHead )
	{
		this->pHead = pHead;
		i.setPos( pHead + (8+1+1+2+12+16+2+5+1+8+8+16+16+8+8)/8, pHead + (8+1+1+2+12)/8 + section_length() - 4 );
	}
	uint16_t service_id() { return UM16((8+1+1+2+12)/8); }

	uint16_t transport_stream_id() { return UM16((8+1+1+2+12+16+2+5+1+8+8)/8); }
	uint16_t original_network_id() { return UM16((8+1+1+2+12+16+2+5+1+8+8+16)/8); }
	uint8_t segment_last_section_number() { return pHead[(8+1+1+2+12+16+2+5+1+8+8+16+16)/8];}
	uint8_t last_table_id() { return pHead[(8+1+1+2+12+16+2+5+1+8+8+16+16+8)/8];}
	// reserv:8
	struct Loop1
	{
		const uint8_t *pHead, *pLimit, *pNext;
		void setPos( const uint8_t* p, const uint8_t* pLimit )
		{
			pHead = p;
			this->pLimit = pLimit;
			pNext = pHead + (16+40+24+3+1+12)/8 + descriptors_loop_length();
			j.setPos( pHead + (16+40+24+3+1+12)/8, pNext );
		}
		void operator ++ ()
		{
			pHead = pNext;
			pNext = pHead + (16+40+24+3+1+12)/8 + descriptors_loop_length();
		}
		uint16_t event_id() { return UM16(0); }
		const uint8_t* start_time() { return &pHead[(16)/8];}
		const uint8_t* duration() { return &pHead[(16+40)/8];}
		uint8_t raw_running_status() { return (pHead[(16+40+24)/8] & 0xe0);}
		uint16_t descriptors_loop_length() { return M16((16+40+24+3+1)/8,12); }

		struct Loop2
		{
			const uint8_t *pHead, *pLimit;
			void setPos( const uint8_t* p, const uint8_t* pLimit )
			{
				pHead = p;
				this->pLimit = pLimit;
			}
			void operator ++ ()
			{
				pHead += 2 + pHead[1];
			}

		} j ;

	} i ;

};

struct Eit_pfInfo {
	int version;
//	int section;
	uint16_t service_id; // PMT program number
	uint16_t event_id;
	QString title;
	QString description;
	QTime start, end;
	vector<uint8_t> descriptor;
};
typedef shared_ptr<Eit_pfInfo> spEit_pfInfo;
typedef map <uint16_t, spEit_pfInfo > Infos;


class Eit_PfInfoLocker : public QObject {
	Q_OBJECT

	QMutex mutex;
	Infos infos;

public:
	Infos& lock() {
		mutex.lock();
		return infos;
	}

	void unlock() {
		mutex.unlock();
	}
};

class PidOperator : public QObject
{
	Q_OBJECT
 protected:

 public:
	virtual void putPacket( const uint8_t* TsPacket ) = 0;
};

class Eit_pfOperator : public PidOperator
{
	Q_OBJECT

	TsPacketReader reader;
	SiBuffer4096 buff;
//	EitData data;
	Conv conv;
	int tsid;

	Eit_PfInfoLocker *infoLocker;
//	BackendThread *pController;
	MainWindow *mainWindow;
 public:
	typedef pair <uint16_t, spEit_pfInfo > InfoPair;

	Eit_pfOperator( MainWindow* p );
	~Eit_pfOperator();

	void reset()
	{
		infoLocker->lock().clear();
		infoLocker->unlock();
		reader.clearBuffer();
	}

	void putPacket( const uint8_t* TsPacket );

signals:
//	void eit_pf_changed( Eit_PfInfoLocker* );
	void eit_pf_changed();
//	void tsid_changed();
};

#undef M16
#undef UM16

#endif // PSISECTION_H
